@@ -7,9 +7,10 @@
 from ._astropy_init import *   # noqa
 # ----------------------------------------------------------------------------
 
+# Define minima for the documentation, but do not bother to explicitly check.
+__minimum_python_version__ = '3.7'
+__minimum_astropy_version__ = '4.0'
+
 if not _ASTROPY_SETUP_:   # noqa
     # For egg_info test builds to pass, put package imports here.
-    pass
-
-
-from .core import file_info, open
+    from .io import file_info, open  # noqa
