@@ -8,17 +8,20 @@ import astropy.units as u
 from astropy.utils import lazyproperty
 
 from ..helpers import sequentialfile as sf
-from ..vlbi_base.base import (make_opener, VLBIFileBase, VLBIFileReaderBase,
-                              VLBIStreamBase,
-                              VLBIStreamReaderBase, VLBIStreamWriterBase)
-from ..vlbi_base.utils import lcm
+from ..base.base import (
+    FileBase,
+    StreamReaderBase, StreamWriterBase,
+    FileOpener, FileInfo)
+from ..base.file_info import FileReaderInfo
+from ..base.utils import lcm
 from .header import DADAHeader
 from .payload import DADAPayload
 from .frame import DADAFrame
 
 
 __all__ = ['DADAFileNameSequencer', 'DADAFileReader', 'DADAFileWriter',
-           'DADAStreamBase', 'DADAStreamReader', 'DADAStreamWriter', 'open']
+           'DADAStreamBase', 'DADAStreamReader', 'DADAStreamWriter',
+           'open', 'info']
 
 
 class DADAFileNameSequencer(sf.FileNameSequencer):
@@ -48,7 +51,7 @@ class DADAFileNameSequencer(sf.FileNameSequencer):
     --------
 
     >>> from baseband import dada
-    >>> dfs = dada.base.DADAFileNameSequencer(
+    >>> dfs = dada.DADAFileNameSequencer(
     ...     '{date}_{file_nr:03d}.dada', {'DATE': "2018-01-01"})
     >>> dfs[10]
     '2018-01-01_010.dada'
@@ -56,7 +59,7 @@ class DADAFileNameSequencer(sf.FileNameSequencer):
     >>> with open(SAMPLE_DADA, 'rb') as fh:
     ...     header = dada.DADAHeader.fromfile(fh)
     >>> template = '{utc_start}.{obs_offset:016d}.000000.dada'
-    >>> dfs = dada.base.DADAFileNameSequencer(template, header)
+    >>> dfs = dada.DADAFileNameSequencer(template, header)
     >>> dfs[0]
     '2013-07-02-01:37:40.0000006400000000.000000.dada'
     >>> dfs[1]
@@ -93,7 +96,7 @@ class DADAFileNameSequencer(sf.FileNameSequencer):
                                         + file_nr * self._file_size)
 
 
-class DADAFileReader(VLBIFileReaderBase):
+class DADAFileReader(FileBase):
     """Simple reader for DADA files.
 
     Wraps a binary filehandle, providing methods to help interpret the data,
@@ -105,6 +108,7 @@ class DADAFileReader(VLBIFileReaderBase):
     fh_raw : filehandle
         Filehandle of the raw binary data file.
     """
+    info = FileReaderInfo()
 
     def read_header(self):
         """Read a single header from the file.
@@ -147,18 +151,16 @@ class DADAFileReader(VLBIFileReaderBase):
         frame_rate : `~astropy.units.Quantity`
             Frames per second.
         """
-        with self.temporary_offset():
-            self.seek(0)
+        with self.temporary_offset(0):
             header = self.read_header()
-            return (header.sample_rate / header.samples_per_frame).to(u.Hz)
+        return (header.sample_rate / header.samples_per_frame).to(u.Hz)
 
 
-class DADAFileWriter(VLBIFileBase):
+class DADAFileWriter(FileBase):
     """Simple writer/mapper for DADA files.
 
-    Adds `write_frame` and `memmap_frame` methods to the VLBI binary file
-    wrapper.  The latter allows one to encode data in pieces, writing to disk
-    as needed.
+    Adds `write_frame` and `memmap_frame` methods to the binary file wrapper.
+    The latter allows one to encode data in pieces, writing to disk as needed.
     """
 
     def write_frame(self, data, header=None, **kwargs):
@@ -206,22 +208,23 @@ class DADAFileWriter(VLBIFileBase):
         return DADAFrame(header, payload)
 
 
-class DADAStreamBase(VLBIStreamBase):
-    """Base for DADA streams."""
+class DADAStreamBase:
+    """Provides sample shape maker and fast index getting/setting."""
 
     _sample_shape_maker = DADAPayload._sample_shape_maker
 
-    def __init__(self, fh_raw, header0, squeeze=True, subset=(), verify=True):
+    def _get_index(self, header):
+        # Override for faster calculation of frame index.
+        return int(round((header['OBS_OFFSET']
+                          - self.header0['OBS_OFFSET'])
+                         / self.header0.payload_nbytes))
 
-        super().__init__(
-            fh_raw=fh_raw, header0=header0, sample_rate=header0.sample_rate,
-            samples_per_frame=header0.samples_per_frame,
-            unsliced_shape=header0.sample_shape, bps=header0.bps,
-            complex_data=header0.complex_data, squeeze=squeeze, subset=subset,
-            fill_value=0., verify=verify)
+    def _set_index(self, header, index):
+        header.update(obs_offset=self.header0['OBS_OFFSET']
+                      + index * self.header0.payload_nbytes)
 
 
-class DADAStreamReader(DADAStreamBase, VLBIStreamReaderBase):
+class DADAStreamReader(DADAStreamBase, StreamReaderBase):
     """DADA format reader.
 
     Allows access to DADA files as a continuous series of samples.
@@ -247,7 +250,7 @@ class DADAStreamReader(DADAStreamBase, VLBIStreamReaderBase):
 
     def __init__(self, fh_raw, squeeze=True, subset=(), verify=True):
         fh_raw = DADAFileReader(fh_raw)
-        header0 = DADAHeader.fromfile(fh_raw)
+        header0 = fh_raw.read_header()
         super().__init__(fh_raw, header0, squeeze=squeeze, subset=subset,
                          verify=verify)
         # Store number of frames, for finding last header.
@@ -263,7 +266,8 @@ class DADAStreamReader(DADAStreamBase, VLBIStreamReaderBase):
                     # If there's only one frame and it's incomplete.
                     if self._nframes == 1:
                         self._header0 = self._last_header
-                        self.samples_per_frame = self.header0.samples_per_frame
+                        self._samples_per_frame = (
+                            self._last_header.samples_per_frame)
                 # Otherwise, ignore the partial frame unless it's the only
                 # frame, in which case raise an EOFError.
                 elif self._nframes == 0:
@@ -327,14 +331,8 @@ class DADAStreamReader(DADAStreamBase, VLBIStreamReaderBase):
         self.fh_raw.seek(0, 2)
         return DADAFrame(self._last_header, last_payload)
 
-    def _tell_frame(self, frame):
-        # Override for faster calculation of frame index.
-        return int(round((frame['OBS_OFFSET']
-                          - self.header0['OBS_OFFSET'])
-                         / self.header0.payload_nbytes))
 
-
-class DADAStreamWriter(DADAStreamBase, VLBIStreamWriterBase):
+class DADAStreamWriter(DADAStreamBase, StreamWriterBase):
     """DADA format writer.
 
     Encodes and writes sequences of samples to file.
@@ -357,21 +355,36 @@ class DADAStreamWriter(DADAStreamBase, VLBIStreamWriterBase):
 
     def _make_frame(self, index):
         header = self.header0.copy()
-        header.update(obs_offset=self.header0['OBS_OFFSET']
-                      + index * self.header0.payload_nbytes)
+        self._set_index(header, index)
         return self.fh_raw.memmap_frame(header)
 
-    def _write_frame(self, frame, valid=True):
+    def _fh_raw_write_frame(self, frame):
         assert frame is self._frame
-        frame.valid = valid
         # Deleting frame flushes memmap'd data to disk.
-        # (Of course, this gets deleted automatically when going out of
-        # scope, and furthermore the link in self._frame will still exist
-        # -- it only gets deleted in VLBIStreamWriter.write)
-        del frame
+        del self._frame
 
 
-opener = make_opener('DADA', globals(), doc="""
+class DADAFileOpener(FileOpener):
+    FileNameSequencer = DADAFileNameSequencer
+
+    def get_fns(self, name, mode, kwargs):
+        fns = super().get_fns(name, mode, kwargs)
+        # For obs_offset we need the first file to know the
+        # actual file_size.
+        if mode[0] == 'r' and 'obs_offset' in name.lower():
+            with io.open(fns[0], 'rb') as fh:
+                header0 = DADAHeader.fromfile(fh)
+            fns = self.FileNameSequencer(name, header0)
+        return fns
+
+    def get_fh(self, name, mode, kwargs):
+        if mode == 'ws' and self.is_sequence(name):
+            kwargs.setdefault('file_size', kwargs['header0'].frame_nbytes)
+
+        return super().get_fh(name, mode, kwargs)
+
+
+open = DADAFileOpener.create(globals(), doc="""
 --- For reading a stream : (see :class:`~baseband.dada.base.DADAStreamReader`)
 
 squeeze : bool, optional
@@ -393,8 +406,8 @@ squeeze : bool, optional
     If `True` (default), writer accepts squeezed arrays as input, and adds
     any dimensions of length unity.
 **kwargs
-    If the header is not given, an attempt will be made to construct one
-    with any further keyword arguments.
+    If no header is given, an attempt is made to construct one from these.
+    For a standard header, this would include the following.
 
 --- Header keywords : (see :meth:`~baseband.dada.DADAHeader.fromvalues`)
 
@@ -429,10 +442,10 @@ Notes
 -----
 For streams, one can also pass to ``name`` a list of files, or a template
 string that can be formatted using 'frame_nr', 'obs_offset', and other header
-keywords (by `~baseband.dada.base.DADAFileNameSequencer`).
+keywords (by `~baseband.dada.DADAFileNameSequencer`).
 
 For writing, one can mimic what is done at quite a few telescopes by using
-the template '{utc_start}_{obs_offset:016d}.000000.dada'.  Unlike for the VLBI
+the template '{utc_start}_{obs_offset:016d}.000000.dada'.  Unlike for most
 openers, ``file_size`` is set to the size of one frame as given by the header.
 
 For reading, to read series such as the above, use something like
@@ -449,63 +462,4 @@ cases it is practically identical to passing in a list or template.
 """)
 
 
-# Need to wrap the opener to be able to deal with file lists or templates.
-def open(name, mode='rs', **kwargs):
-    # Extract needed kwargs (and keep some from being passed to opener).
-    header0 = kwargs.get('header0', None)
-
-    # Check if ``name`` is a template or sequence.
-    is_template = isinstance(name, str) and ('{' in name and '}' in name)
-    is_sequence = isinstance(name, (tuple, list, sf.FileNameSequencer))
-
-    # For stream writing, header0 is needed; for reading, it is needed for
-    # initializing a template only.
-    if 'b' not in mode:
-        # Initialize header0 if it doesn't yet exist.
-        if header0 is None:
-            if 'w' in mode:
-                # Store squeeze.
-                passed_kwargs = ({'squeeze': kwargs.pop('squeeze')}
-                                 if 'squeeze' in kwargs.keys() else {})
-                # Make header0.
-                header0 = DADAHeader.fromvalues(**kwargs)
-                # Pass squeeze and header0 on to stream writer.
-                kwargs = passed_kwargs
-                kwargs['header0'] = header0
-
-            elif is_template:
-                # Store parameters to pass.
-                passed_kwargs = {key: kwargs.pop(key) for key in
-                                 ('squeeze', 'subset', 'verify')
-                                 if key in kwargs}
-                kwargs = {key.upper(): value for key, value in kwargs.items()}
-
-                # If obs_offset is needed, make a temporary file sequence to
-                # read in header0.
-                if ('OBS_OFFSET' in name or 'obs_offset' in name):
-                    for key in ('OBS_OFFSET', 'FILE_SIZE'):
-                        kwargs.setdefault(key, 0)
-                    first_file = DADAFileNameSequencer(name, kwargs)[0]
-                    with io.open(first_file, 'rb') as fh:
-                        header0 = DADAHeader.fromfile(fh)
-                # If obs_offset isn't needed, just use the kwargs.
-                else:
-                    header0 = kwargs
-
-                kwargs = passed_kwargs
-
-        if is_template:
-            name = DADAFileNameSequencer(name, header0)
-
-    # If writing with a template or sequence, pass ``file_size``.
-    if 'w' in mode and (is_template or is_sequence):
-        if 'b' in mode:
-            raise ValueError("does not support opening a file sequence in "
-                             "'wb' mode.  Try passing in a SequentialFile "
-                             "object instead.")
-        kwargs['file_size'] = header0.frame_nbytes
-
-    return opener(name, mode, **kwargs)
-
-
-open.__doc__ = opener.__doc__
+info = FileInfo.create(globals())
