@@ -1,5 +1,6 @@
 # Licensed under the GPLv3 - see LICENSE
 import copy
+import pickle
 
 import pytest
 import numpy as np
@@ -12,16 +13,16 @@ from ...data import SAMPLE_PUPPI as SAMPLE_FILE
 
 
 class TestGUPPI:
-    def setup(self):
+    def setup_class(cls):
         with open(SAMPLE_FILE, 'rb') as fh:
-            self.header = guppi.GUPPIHeader.fromfile(fh)
-            self.payload = guppi.GUPPIPayload.fromfile(fh, self.header,
-                                                       memmap=False)
+            cls.header = guppi.GUPPIHeader.fromfile(fh)
+            cls.payload = guppi.GUPPIPayload.fromfile(fh, cls.header,
+                                                      memmap=False)
         # Create a header with no overlap for stream writers.
-        self.header_w = self.header.copy()
-        self.header_w.overlap = 0
-        self.header_w.payload_nbytes = self.header.payload_nbytes - (
-            self.header._bpcs * self.header.overlap // 8)
+        cls.header_w = cls.header.copy()
+        cls.header_w.overlap = 0
+        cls.header_w.payload_nbytes = cls.header.payload_nbytes - (
+            cls.header._bpcs * cls.header.overlap // 8)
 
     def test_header(self, tmpdir):
         with open(SAMPLE_FILE, 'rb') as fh:
@@ -172,6 +173,25 @@ class TestGUPPI:
             guppi.GUPPIHeader.fromvalues(nchan=1, npol=1, bps=4,
                                          samples_per_frame=10001)
 
+    def test_header_comment_cards(self, tmpdir):
+        # Actually not obvious GUPPI itself allows it, but we might as
+        # well test that we do not fail in setting comments.
+        with open(SAMPLE_FILE, 'rb') as fh:
+            header = guppi.GUPPIHeader.fromfile(fh)
+        assert 'OBSNCHAN' not in header.comments
+        header1 = header.copy()
+        header1['OBSNCHAN'] = header['OBSNCHAN'], 'number of channels'
+        assert header1.comments['OBSNCHAN'] == 'number of channels'
+        name = str(tmpdir.join('guppi_header.test'))
+        with open(name, 'wb') as fw:
+            header1.tofile(fw)
+
+        with open(name, 'rb') as fr:
+            header2 = guppi.GUPPIHeader.fromfile(fr)
+
+        assert header2 == header
+        assert header2.comments['OBSNCHAN'] == 'number of channels'
+
     def test_payload(self, tmpdir):
         payload = self.payload
         assert payload.nbytes == 16384
@@ -211,7 +231,7 @@ class TestGUPPI:
             with pytest.raises(EOFError):
                 # Too few bytes.
                 s.seek(100)
-                guppi.GUPPIPayload.fromfile(s, self.header)
+                guppi.GUPPIPayload.fromfile(s, self.header, memmap=False)
 
         payload3 = guppi.GUPPIPayload.fromdata(payload.data, bps=8)
         assert payload3 == payload
@@ -252,6 +272,9 @@ class TestGUPPI:
         item = (slice(547, 829, 2), slice(None), np.array([2, 1]))
         assert np.all(payload_tfirst[item] == payload[item])
 
+        with pytest.raises(ValueError, match='cannot encode'):
+            guppi.GUPPIPayload.fromdata(payload.data, bps=4)
+
     def test_file_reader(self):
         with guppi.open(SAMPLE_FILE, 'rb') as fh:
             header = fh.read_header()
@@ -274,6 +297,7 @@ class TestGUPPI:
             assert info.sample_shape == header.sample_shape
             assert info.start_time == header.start_time
             assert info.samples_per_frame == header.samples_per_frame
+            assert info.overlap == header.overlap
             # Note: sample_rate used to be wrong, since calculated from
             # from frame rate and samples_per_frame w/o corr. for overlap.
             assert info.sample_rate == header.sample_rate
@@ -351,13 +375,18 @@ class TestGUPPI:
             frame3 = fw.read_frame()
 
         assert frame3 == frame
+
         # Now memmap file to be written to.
-        with guppi.open(filename, 'wb') as fw:
+        # Use a new file since the mmap used in frame3 may still have a handle
+        # to the previous one, which causes (sensible) failures on windows.
+        # See https://bugs.python.org/issue40720
+        filename2 = str(tmpdir.join('testguppi2.raw'))
+        with guppi.open(filename2, 'wb') as fw:
             frame4 = fw.memmap_frame(frame.header)
         # Initially no data set, so frames should not match yet.
         assert frame4 != frame
         # So, if we read this file, it also should not match.
-        with guppi.open(filename, 'rb') as fw:
+        with guppi.open(filename2, 'rb') as fw:
             frame5 = fw.read_frame()
         assert frame5 != frame
 
@@ -370,33 +399,36 @@ class TestGUPPI:
         assert frame4 == frame
         # Delete to flush to disk just to be sure, then read and check it.
         del frame4
-        with guppi.open(filename, 'rb') as fn:
+        with guppi.open(filename2, 'rb') as fn:
             frame6 = fn.read_frame()
 
         assert frame6 == frame
         # Some further tests for completeness;
         # initiate frame using data and header keywords.
-        with guppi.open(filename, 'wb') as fw:
+        filename3 = str(tmpdir.join('testguppi3.raw'))
+        with guppi.open(filename3, 'wb') as fw:
             fw.write_frame(self.payload.data, **self.header)
-        with guppi.open(filename, 'rb') as fh:
+        with guppi.open(filename3, 'rb') as fh:
             frame7 = fh.read_frame()
         assert frame7 == frame
         # memmap frame using header keywords.
-        with guppi.open(filename, 'wb') as fw:
+        filename4 = str(tmpdir.join('testguppi4.raw'))
+        with guppi.open(filename4, 'wb') as fw:
             frame8 = fw.memmap_frame(**self.header)
             frame8[:] = self.payload.data
         assert frame8 == frame
         del frame8
-        with guppi.open(filename, 'rb') as fh:
+        with guppi.open(filename4, 'rb') as fh:
             frame9 = fh.read_frame()
         assert frame9 == frame
 
     def test_filestreamer(self, tmpdir):
         start_time = self.header.time
+        nsample = 4*1024 - 3*64
         with guppi.open(SAMPLE_FILE, 'rs') as fh:
             assert fh.header0 == self.header
             assert fh.sample_shape == (2, 4)
-            assert fh.shape == (3840,) + fh.sample_shape
+            assert fh.shape == (nsample,) + fh.sample_shape
             assert fh.size == np.prod(fh.shape)
             assert fh.ndim == len(fh.shape)
             assert fh.start_time == start_time
@@ -417,7 +449,7 @@ class TestGUPPI:
             fh.seek(fh.start_time + 100 / (250 * u.Hz))
             assert fh.tell() == 100
             assert (np.abs(fh.stop_time
-                           - (start_time + 3840 / (250 * u.Hz))) < 1. * u.ns)
+                           - (start_time + nsample / (250 * u.Hz))) < 1.*u.ns)
             fh.seek(1, 'end')
             with pytest.raises(EOFError):
                 fh.read()
@@ -458,7 +490,8 @@ class TestGUPPI:
         # initialisation by header keywords.  This also tests writing
         # with squeeze=True.
         h = self.header
-        with guppi.open(filename, 'ws', time=h.time, bps=h.bps,
+        filename2 = str(tmpdir.join('testguppi2.raw'))
+        with guppi.open(filename2, 'ws', time=h.time, bps=h.bps,
                         sample_rate=h.sample_rate,
                         pktsize=h['PKTSIZE'], overlap=0,
                         payload_nbytes=self.header_w.payload_nbytes // 4,
@@ -468,7 +501,7 @@ class TestGUPPI:
             assert (np.abs(fw.time - (start_time + spf / (250 * u.Hz)))
                     < 1. * u.ns)
 
-        with guppi.open(filename, 'rs') as fh:
+        with guppi.open(filename2, 'rs') as fh:
             data_onepol = fh.read()
             assert np.abs(fh.start_time - start_time) < 1.*u.ns
             assert np.abs(fh.stop_time
@@ -489,7 +522,7 @@ class TestGUPPI:
             assert np.all(data_sub[:, 1] == record[:, 1, 0])
 
         # Test that squeeze attribute works on read (including in-place read).
-        with guppi.open(filename, 'rs', squeeze=False) as fh:
+        with guppi.open(filename2, 'rs', squeeze=False) as fh:
             assert fh.sample_shape == (1, 2)
             assert fh.sample_shape.npol == 1
             assert fh.sample_shape.nchan == 2
@@ -503,7 +536,8 @@ class TestGUPPI:
 
         # Test that squeeze=False works on write (we checked above it works
         # with squeeze=True).
-        with guppi.open(filename, 'ws', time=h.time, bps=h.bps,
+        filename3 = str(tmpdir.join('testguppi3.raw'))
+        with guppi.open(filename3, 'ws', time=h.time, bps=h.bps,
                         sample_rate=h.sample_rate,
                         pktsize=h['PKTSIZE'], overlap=0,
                         payload_nbytes=self.header_w.payload_nbytes // 4,
@@ -513,10 +547,23 @@ class TestGUPPI:
             assert (np.abs(fw.time - (start_time + spf / (250 * u.Hz)))
                     < 1. * u.ns)
 
-        with guppi.open(filename, 'rs', squeeze=False) as fh:
+        with guppi.open(filename3, 'rs', squeeze=False) as fh:
             data_onepol = fh.read()
         assert np.all(data_onepol.squeeze() == self.payload[:spf, 0, :2])
 
+    def test_stream_overlap(self):
+        # See gh-264
+        with guppi.open(SAMPLE_FILE, 'rs') as fh:
+            overlap = fh.header0.overlap
+            # Check "normal" end.
+            fh.seek(4*fh.samples_per_frame)
+            data = fh.read()
+            assert len(data) == overlap
+            fh.seek(-1, 2)
+            assert fh.tell() == 4*fh.samples_per_frame + overlap - 1
+            data = fh.read()
+            assert len(data) == 1
+
     def test_incomplete_stream(self, tmpdir):
         """Test that writing an incomplete stream is possible, and that frame
         set is valid but invalid samples use the fill value.
@@ -532,10 +579,32 @@ class TestGUPPI:
             assert np.all(data[:10] == self.payload[:10])
             assert np.all(data[10:] == fwr.fill_value)
 
+    def test_pickle(self):
+        # Only simple tests here; more complete ones in vdif.
+        with guppi.open(SAMPLE_FILE, 'rs') as fh:
+            fh.seek(6)
+            pickled = pickle.dumps(fh)
+            fh.read(3)
+            with pickle.loads(pickled) as fh2:
+                assert fh2.tell() == 6
+                fh2.read(10)
+
+            assert fh.tell() == 9
+
+        with pickle.loads(pickled) as fh3:
+            assert fh3.tell() == 6
+            fh3.read(1)
+
+        closed = pickle.dumps(fh)
+        with pickle.loads(closed) as fh4:
+            assert fh4.closed
+            with pytest.raises(ValueError):
+                fh4.read(1)
+
     def test_multiple_files_stream(self, tmpdir):
         start_time = self.header_w.time
         with guppi.open(SAMPLE_FILE, 'rs') as fh:
-            data = fh.read()
+            data = fh.read(3840)  # omit overlap for test
         filenames = (str(tmpdir.join('guppi_1.raw')),
                      str(tmpdir.join('guppi_2.raw')))
         with guppi.open(filenames, 'ws', header0=self.header_w,
@@ -567,6 +636,8 @@ class TestGUPPI:
         assert np.all(data2 == data)
 
         # Pass sequentialfile objects to reader.
+        filenames = (str(tmpdir.join('guppi2_1.raw')),
+                     str(tmpdir.join('guppi2_2.raw')))
         with sf.open(filenames, 'w+b',
                      file_size=2*self.header_w.frame_nbytes) as fraw, \
                 guppi.open(fraw, 'ws', header0=self.header_w) as fw:
@@ -575,8 +646,16 @@ class TestGUPPI:
         with sf.open(filenames, 'rb') as fraw, \
                 guppi.open(fraw, 'rs') as fr:
             data3 = fr.read()
+            # Test pickling in the process
+            pickled = pickle.dumps(fr)
         assert np.all(data3 == data)
 
+        with pickle.loads(pickled) as fr2:
+            assert fr2.tell() == fr2.shape[0]
+            fr2.seek(-10, 2)
+            datap = fr2.read()
+        assert np.all(datap.squeeze() == data[-10:])
+
         # Check that we can't pass a filename sequence in 'wb' mode.
         with pytest.raises(ValueError):
             guppi.open(filenames, 'wb')
@@ -590,18 +669,19 @@ class TestGUPPI:
         # Try reading a file with an incomplete payload.
         with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'wb') as fw:
             fw.write(puppi_raw[:len(puppi_raw) - 6091])
+        nsample = 3*1024 - 2*64  # 3 frames minus 2 overlaps
         with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'rs') as fn:
-            assert fn.shape == (2880, 2, 4)
+            assert fn.shape == (nsample, 2, 4)
             assert np.abs(fn.stop_time
-                          - fn.start_time - 2880 / (250 * u.Hz)) < 1. * u.ns
+                          - fn.start_time - nsample / (250 * u.Hz)) < 1. * u.ns
 
         # Try reading a file with an incomplete header.
         with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'wb') as fw:
             fw.write(puppi_raw[:len(puppi_raw) - 17605])
         with guppi.open(str(tmpdir.join('puppi_partframe.raw')), 'rs') as fn:
-            assert fn.shape == (2880, 2, 4)
+            assert fn.shape == (nsample, 2, 4)
             assert np.abs(fn.stop_time
-                          - fn.start_time - 2880 / (250 * u.Hz)) < 1. * u.ns
+                          - fn.start_time - nsample / (250 * u.Hz)) < 1. * u.ns
 
     def test_chan_ordered_stream(self, tmpdir):
         """Test encoding and decoding frames and streams that use
@@ -610,7 +690,7 @@ class TestGUPPI:
         filename = str(tmpdir.join('testguppi.raw'))
 
         with guppi.open(SAMPLE_FILE) as fh:
-            data = fh.read()
+            data = fh.read(3840)  # omit overlap for test
 
         header = self.header.copy()
         header.channels_first = False
@@ -626,7 +706,7 @@ class TestGUPPI:
     def test_template_stream(self, tmpdir):
         start_time = self.header_w.time
         with guppi.open(SAMPLE_FILE, 'rs') as fh:
-            data = fh.read()
+            data = fh.read(3840)  # omit overlap for test
 
         # Simple template with file number counter.
         template = str(tmpdir.join('guppi_{file_nr:02d}.raw'))
