@@ -12,12 +12,14 @@ import numpy as np
 import astropy.units as u
 from astropy.time import Time, TimeDelta
 
-from ..vlbi_base.header import (four_word_struct, eight_word_struct,
-                                fixedvalue, HeaderParser, VLBIHeaderBase)
+from ..base.header import (four_word_struct, eight_word_struct,
+                           HeaderParser, VLBIHeaderBase)
+from ..base.utils import fixedvalue
 from ..mark5b.header import Mark5BHeader
 
 
-__all__ = ['VDIFHeader', 'VDIFBaseHeader', 'VDIFSampleRateHeader',
+__all__ = ['VDIFHeader', 'VDIFBaseHeader',
+           'VDIFNoSampleRateHeader', 'VDIFSampleRateHeader',
            'VDIFLegacyHeader', 'VDIFHeader0', 'VDIFHeader1',
            'VDIFHeader2', 'VDIFHeader3', 'VDIFMark5BHeader',
            'VDIF_HEADER_CLASSES']
@@ -50,7 +52,7 @@ class VDIFHeaderMeta(type):
 
         # Ignore VDIFHeader and VDIFBaseHeader, register others
         if name not in ('VDIFHeader', 'VDIFBaseHeader',
-                        'VDIFSampleRateHeader'):
+                        'VDIFNoSampleRateHeader', 'VDIFSampleRateHeader'):
 
             # Extract edv from class; convert to -1 if edv is False
             # for VDIFLegacy
@@ -112,14 +114,15 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
     # are broken for some threads of the EVN 2014 data, so apparently
     # that is not a safe bet.
 
-    _properties = ('frame_nbytes', 'payload_nbytes', 'bps', 'nchan',
-                   'samples_per_frame', 'station', 'time')
+    _properties = ('frame_nbytes', 'payload_nbytes', 'bps', 'complex_data',
+                   'nchan', 'sample_shape', 'samples_per_frame',
+                   'station', 'ref_time', 'time')
     """Properties accessible/usable in initialisation for all VDIF headers."""
 
     _edv = None
     _struct = eight_word_struct
 
-    def __new__(cls, words, edv=None, verify=True, **kwargs):
+    def __new__(cls, words=None, edv=None, verify=True, **kwargs):
         # We use edv to define which class we return.
         if edv is None:
             # If not given, try the class default.
@@ -214,37 +217,15 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
         Note that to set ``time`` to non-integer seconds one also needs to
         pass in ``frame_rate`` or ``sample_rate``.
         """
-        # TODO: here, we act differently depending on what information the
-        # VDIF header subclass can handle. Ideally, these different actions
-        # would be stored on the subclasses themselves, but then their super
-        # calls would end up here anyway.  One could by-pass those by checking
-        # whether cls is VDIFHeader.  But another issue is that the headers
-        # with a sample rate are currently subclasses of those without, so
-        # the super chain doesn't work well.
-        #
         # Some defaults that are not done by setting properties.
         kwargs.setdefault('legacy_mode', edv is False)
         if edv is not False:
             kwargs['edv'] = edv
-        # For setting time, one normally needs a frame rate.  If headers
-        # provide this information, we can just proceed.
-        if 'time' not in kwargs or 'frame_rate' in VDIF_HEADER_CLASSES.get(
-                edv if edv is not False else -1, VDIFBaseHeader)._properties:
-            return super().fromvalues(edv, verify=verify, **kwargs)
-        # If the VDIF header subclass does not provide the frame rate, we
-        # first initialize without time, and then set the time explicitly
-        # using whatever frame_rate or sample_rate was passed.
-        time = kwargs.pop('time')
-        sample_rate = kwargs.pop('sample_rate', None)
-        frame_rate = kwargs.pop('frame_rate', None)
-        # Skip verification in super as we do it after time is set.
-        self = super().fromvalues(edv, verify=False, **kwargs)
-        if frame_rate is None and sample_rate is not None:
-            frame_rate = sample_rate / self.samples_per_frame
-        self.set_time(time, frame_rate=frame_rate)
-        if verify:
-            self.verify()
-        return self
+
+        if 'time' in kwargs and 'ref_epoch' not in kwargs:
+            kwargs.setdefault('ref_time', kwargs['time'])
+
+        return super().fromvalues(edv, verify=verify, **kwargs)
 
     @classmethod
     def fromkeys(cls, **kwargs):
@@ -342,6 +323,15 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
                              "is only possible for single-channel data.")
         self['bits_per_sample'] = int(bps) - 1
 
+    @property
+    def complex_data(self):
+        """Whether the data are complex."""
+        return self['complex_data']
+
+    @complex_data.setter
+    def complex_data(self, complex_data):
+        self['complex_data'] = complex_data
+
     @property
     def nchan(self):
         """Number of channels in the frame."""
@@ -356,6 +346,15 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
                              "that is a power of two.")
         self['lg2_nchan'] = int(lg2_nchan)
 
+    @property
+    def sample_shape(self):
+        """Shape of a sample in the payload (nchan,)."""
+        return (self.nchan,)
+
+    @sample_shape.setter
+    def sample_shape(self, sample_shape):
+        self.nchan, = sample_shape
+
     @property
     def samples_per_frame(self):
         """Number of complete samples in the frame."""
@@ -398,6 +397,20 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
         assert int(station_id) == station_id
         self['station_id'] = station_id
 
+    @property
+    def ref_time(self):
+        """Reference time for the header, as encoded by 'ref_epoch'.
+
+        If set, will use the nearest reference epoch before the given time.
+        """
+        return ref_epochs[self['ref_epoch']]
+
+    @ref_time.setter
+    def ref_time(self, ref_time):
+        assert ref_time > ref_epochs[0]
+        ref_index = np.searchsorted((ref_epochs - ref_time).sec, 0) - 1
+        self['ref_epoch'] = ref_index
+
     def get_time(self, frame_rate=None):
         """Converts ref_epoch, seconds, and frame_nr to Time object.
 
@@ -426,7 +439,7 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
 
             offset = (frame_nr / frame_rate).to_value(u.s)
 
-        return (ref_epochs[self['ref_epoch']]
+        return (self.ref_time
                 + TimeDelta(self['seconds'], offset, format='sec'))
 
     def set_time(self, time, frame_rate=None):
@@ -442,10 +455,7 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
         frame_rate : `~astropy.units.Quantity`, optional
             For calculating 'frame_nr' from the fractional seconds.
         """
-        assert time > ref_epochs[0]
-        ref_index = np.searchsorted((ref_epochs - time).sec, 0) - 1
-        self['ref_epoch'] = ref_index
-        seconds = time - ref_epochs[ref_index]
+        seconds = time - self.ref_time
         int_sec = int(seconds.sec)
 
         # Round to nearest ns to handle timestamp difference errors.
@@ -471,7 +481,44 @@ class VDIFHeader(VLBIHeaderBase, metaclass=VDIFHeaderMeta):
     time = property(get_time, set_time)
 
 
-class VDIFLegacyHeader(VDIFHeader):
+class VDIFNoSampleRateHeader(VDIFHeader):
+    def update(self, *, time=None, frame_rate=None, sample_rate=None,
+               verify=True, **kwargs):
+        """Update the header by setting keywords or properties.
+
+        Here, any keywords matching header keys are applied first, and any
+        remaining ones are used to set header properties, in the order set
+        by the class (in ``_properties``).
+
+        Parameters
+        ----------
+        time : `~astropy.time.Time`, optional
+            A possible new time.  This is updated last.
+        frame_rate : `~astropy.units.Quantity`, optional
+            The frame rate to use in calculating the frame number from the
+            time.  Needed for times at non-integer seconds.  Ignored if no
+            ``time`` is given.
+        sample_rate : `~astropy.units.Quantity`, optional
+            Alternative to ``frame_rate``.  Ignored if no ``time`` is given,
+            or if ``frame_rate`` is given.
+        verify : bool, optional
+            If `True` (default), verify integrity after updating.
+        **kwargs
+            Arguments used to set keywords and properties.
+        """
+        super().update(verify=False, **kwargs)
+
+        if time is not None:
+            if frame_rate is None and sample_rate is not None:
+                frame_rate = sample_rate / self.samples_per_frame
+
+            self.set_time(time, frame_rate=frame_rate)
+
+        if verify:
+            self.verify()
+
+
+class VDIFLegacyHeader(VDIFNoSampleRateHeader):
     """Legacy VDIF header that uses only 4 32-bit words.
 
     See Section 6 of
@@ -507,7 +554,7 @@ class VDIFLegacyHeader(VDIFHeader):
 class VDIFBaseHeader(VDIFHeader):
     """Base for non-legacy VDIF headers that use 8 32-bit words."""
 
-    _header_parser = VDIFLegacyHeader._header_parser + HeaderParser(
+    _header_parser = VDIFLegacyHeader._header_parser | HeaderParser(
         (('legacy_mode', (0, 30, 1, False)),  # Repeat, to change default.
          ('edv', (4, 24, 8))))
 
@@ -530,7 +577,7 @@ class VDIFBaseHeader(VDIFHeader):
             assert self['sync_pattern'] == self._sync_pattern
 
 
-class VDIFHeader0(VDIFBaseHeader):
+class VDIFHeader0(VDIFBaseHeader, VDIFNoSampleRateHeader):
     """VDIF Header for EDV=0.
 
     EDV=0 implies the extended user data fields are not used.
@@ -545,7 +592,7 @@ class VDIFHeader0(VDIFBaseHeader):
 class VDIFSampleRateHeader(VDIFBaseHeader):
     """Base for VDIF headers that include the sample rate (EDV= 1, 3, 4)."""
 
-    _header_parser = VDIFBaseHeader._header_parser + HeaderParser(
+    _header_parser = VDIFBaseHeader._header_parser | HeaderParser(
         (('sampling_unit', (4, 23, 1)),
          ('sampling_rate', (4, 0, 23)),
          ('sync_pattern', (5, 0, 32, 0xACABFEED))))
@@ -651,7 +698,7 @@ class VDIFHeader1(VDIFSampleRateHeader):
     See https://vlbi.org/wp-content/uploads/2019/03/vdif_extension_0x01.pdf
     """
     _edv = 1
-    _header_parser = VDIFSampleRateHeader._header_parser + HeaderParser(
+    _header_parser = VDIFSampleRateHeader._header_parser | HeaderParser(
         (('das_id', (6, 0, 64, 0x0)),))
 
     _invariants = (VDIFSampleRateHeader._invariants
@@ -665,7 +712,7 @@ class VDIFHeader3(VDIFSampleRateHeader):
     See https://vlbi.org/wp-content/uploads/2019/03/vdif_extension_0x03.pdf
     """
     _edv = 3
-    _header_parser = VDIFSampleRateHeader._header_parser + HeaderParser(
+    _header_parser = VDIFSampleRateHeader._header_parser | HeaderParser(
         (('frame_length', (2, 0, 24, 629)),  # Repeat, to set default.
          ('loif_tuning', (6, 0, 32, 0x0)),
          ('_7_28_4', (7, 28, 4, 0x0)),
@@ -698,7 +745,7 @@ class VDIFHeader3(VDIFSampleRateHeader):
         return cls.nbytes + cls.payload_nbytes
 
 
-class VDIFHeader2(VDIFBaseHeader):
+class VDIFHeader2(VDIFBaseHeader, VDIFNoSampleRateHeader):
     """VDIF Header for EDV=2.
 
     See https://vlbi.org/wp-content/uploads/2019/03/alma-vdif-edv.pdf
@@ -710,7 +757,7 @@ class VDIFHeader2(VDIFBaseHeader):
     """
     _edv = 2
 
-    _header_parser = VDIFBaseHeader._header_parser + HeaderParser(
+    _header_parser = VDIFBaseHeader._header_parser | HeaderParser(
         (('complex_data', (3, 31, 1, 0x0)),  # Repeat, to set default.
          ('bits_per_sample', (3, 26, 5, 0x1)),  # Repeat, to set default.
          ('pol', (4, 0, 1)),
@@ -733,7 +780,7 @@ class VDIFHeader2(VDIFBaseHeader):
         assert self.bps == 2 and not self['complex_data']
 
 
-class VDIFMark5BHeader(VDIFBaseHeader, Mark5BHeader):
+class VDIFMark5BHeader(VDIFBaseHeader, VDIFNoSampleRateHeader, Mark5BHeader):
     """Mark 5B over VDIF (EDV=0xab).
 
     See https://vlbi.org/wp-content/uploads/2019/03/vdif_extension_0xab.pdf
@@ -741,8 +788,8 @@ class VDIFMark5BHeader(VDIFBaseHeader, Mark5BHeader):
     _edv = 0xab
     # Repeat 'frame_length' to set default.
     _header_parser = (VDIFBaseHeader._header_parser
-                      + HeaderParser((('frame_length', (2, 0, 24, 1254)),))
-                      + HeaderParser(tuple(
+                      | HeaderParser((('frame_length', (2, 0, 24, 1254)),))
+                      | HeaderParser(tuple(
                           ((k if k != 'frame_nr' else 'mark5b_frame_nr'),
                            (v[0] + 4,) + v[1:])
                           for (k, v) in Mark5BHeader._header_parser.items())))
@@ -760,6 +807,7 @@ class VDIFMark5BHeader(VDIFBaseHeader, Mark5BHeader):
         super().verify()
         assert self['frame_length'] == 1254  # payload+header=10000+32 bytes/8
         assert self['frame_nr'] == self['mark5b_frame_nr']
+        assert not self['complex_data']
         # Check consistency of time down to the second (since some Mark 5B
         # headers do not store 'bcd_fraction').
         day, seconds = divmod(self['seconds'], 86400)
@@ -775,10 +823,18 @@ class VDIFMark5BHeader(VDIFBaseHeader, Mark5BHeader):
     def frame_nbytes(cls):
         return cls.nbytes + cls.payload_nbytes
 
+    @fixedvalue
+    def complex_data(cls):
+        return False
+
     def __setitem__(self, item, value):
-        super().__setitem__(item, value)
-        if item == 'frame_nr':
-            super().__setitem__('mark5b_frame_nr', value)
+        if item == 'complex_data':
+            # Pass by fixed-value setter - can only set to False.
+            self.complex_data = value
+        else:
+            super().__setitem__(item, value)
+            if item == 'frame_nr':
+                super().__setitem__('mark5b_frame_nr', value)
 
     def get_time(self, frame_rate=None):
         """
