@@ -1,12 +1,13 @@
 # Licensed under the GPLv3 - see LICENSE
+import pickle
+
 import pytest
 import numpy as np
 from astropy.time import Time
 import astropy.units as u
 
-from ... import vdif, vlbi_base
-from ...vlbi_base.base import HeaderNotFoundError
-from ...helpers import sequentialfile as sf
+from ... import vdif, base
+from ...base.base import HeaderNotFoundError
 from ...data import (SAMPLE_VDIF as SAMPLE_FILE, SAMPLE_VLBI_VDIF as
                      SAMPLE_VLBI, SAMPLE_MWA_VDIF as SAMPLE_MWA,
                      SAMPLE_AROCHIME_VDIF as SAMPLE_AROCHIME,
@@ -50,13 +51,16 @@ class TestVDIF:
         mjd, frac = divmod(header.time.mjd, 1)
         assert mjd == 56824
         assert round(frac * 86400) == 21367
+        assert header.ref_time == Time('2014-01-01')
         assert header.payload_nbytes == 5000
         assert header.frame_nbytes == 5032
         assert header['thread_id'] == 1
         assert header.sample_rate == 32*u.MHz
         assert header.samples_per_frame == 20000
         assert header.nchan == 1
+        assert header.sample_shape == (1,)
         assert header.bps == 2
+        assert not header.complex_data
         assert not header['complex_data']
         assert header.mutable is False
         with open(str(tmpdir.join('test.vdif')), 'w+b') as s:
@@ -90,11 +94,11 @@ class TestVDIF:
             sideband=header['sideband'], major_rev=header['major_rev'],
             minor_rev=header['minor_rev'], personality=header['personality'],
             _7_28_4=header['_7_28_4'])
-        # The same header, but created by passing time and sample rate.
+        # The same header, but created by passing time and frame rate.
         header4_usetime = vdif.VDIFHeader.fromvalues(
             edv=header.edv, time=header.time,
             samples_per_frame=header.samples_per_frame,
-            station=header.station, sample_rate=header.sample_rate,
+            station=header.station, frame_rate=header.frame_rate,
             bps=header.bps, complex_data=header['complex_data'],
             thread_id=header['thread_id'],
             loif_tuning=header['loif_tuning'], dbe_unit=header['dbe_unit'],
@@ -137,26 +141,39 @@ class TestVDIF:
 
         # Test initializing EDV 0 with properties, but off of 1 second mark so
         # frame_nr is used.
-        header8 = vdif.VDIFHeader.fromvalues(
+        header7 = vdif.VDIFHeader.fromvalues(
             edv=0, ref_epoch=headerT['ref_epoch'], seconds=headerT['seconds'],
-            frame_nr=headerT['frame_nr'], complex_data=headerT['complex_data'],
+            frame_nr=headerT['frame_nr'], complex_data=headerT.complex_data,
             samples_per_frame=headerT.samples_per_frame, bps=headerT.bps,
             station=headerT.station, thread_id=headerT['thread_id'])
-        assert header8['ref_epoch'] == headerT['ref_epoch']
-        assert header8['seconds'] == headerT['seconds']
-        assert header8['frame_nr'] == headerT['frame_nr']
+        assert header7['ref_epoch'] == headerT['ref_epoch']
+        assert header7['seconds'] == headerT['seconds']
+        assert header7['frame_nr'] == headerT['frame_nr']
         # The same header, but created by passing time and sample rate.
-        header8_usetime = vdif.VDIFHeader.fromvalues(
+        header7_usetime = vdif.VDIFHeader.fromvalues(
             edv=0, time=headerT.time, sample_rate=headerT.sample_rate,
-            complex_data=headerT['complex_data'], bps=headerT.bps,
+            complex_data=headerT.complex_data, bps=headerT.bps,
             samples_per_frame=headerT.samples_per_frame,
             station=headerT.station, thread_id=headerT['thread_id'])
-        assert header8_usetime == header8
+        assert header7_usetime == header7
+
+        # Finally, check that we can pass in a different reference time, but
+        # still get the right time and frame number.
+        header8 = vdif.VDIFHeader.fromvalues(
+            edv=0, ref_epoch=0, time=headerT.time,
+            sample_rate=headerT.sample_rate,
+            complex_data=headerT.complex_data, bps=headerT.bps,
+            samples_per_frame=headerT.samples_per_frame,
+            station=headerT.station, thread_id=headerT['thread_id'])
+        assert header8.ref_time == Time('2000-01-01')
+        assert abs(header8.get_time(frame_rate=headerT.frame_rate)
+                   - headerT.time) < 1.*u.ns
+        assert header8['frame_nr'] == headerT['frame_nr']
 
         # Without a sample rate or frame_nr, cannot initialize using time.
         with pytest.raises(ValueError):
             vdif.VDIFHeader.fromvalues(
-                edv=0, time=headerT.time, complex_data=headerT['complex_data'],
+                edv=0, time=headerT.time, complex_data=headerT.complex_data,
                 bps=headerT.bps, samples_per_frame=headerT.samples_per_frame,
                 station=headerT.station, thread_id=headerT['thread_id'])
 
@@ -165,12 +182,17 @@ class TestVDIF:
             vdif.VDIFHeader.fromvalues(
                 edv=1, time=headerT.time, station=headerT.station,
                 samples_per_frame=headerT.samples_per_frame,
-                bps=headerT.bps, complex_data=headerT['complex_data'],
+                bps=headerT.bps, complex_data=headerT.complex_data,
                 thread_id=headerT['thread_id'])
 
-        # Check rounding in corner case.
+        # Check rounding in corner case, both with and without changing
+        # the reference time as well.
         header9 = headerT.copy()
-        header9.set_time(Time('2018-01-01T00:34:07.999999999996'))
+        time = Time('2018-01-01T00:34:07.999999999996')
+        header9.time = time
+        assert header9['seconds'] == 126232450
+        assert header9['frame_nr'] == 0
+        header9.ref_time = header9.time = time
         assert header9['seconds'] == 2048
         assert header9['frame_nr'] == 0
 
@@ -191,12 +213,12 @@ class TestVDIF:
             header11.nchan = 2
         header11.bps = 8
         assert header11.bps == 8
-        header11.nchan = 2
+        header11.sample_shape = (2,)
         assert header11.nchan == 2
 
         header12 = header.copy()
         header12.nchan = 2
-        assert header12.nchan == 2
+        assert header12.sample_shape == (2,)
         with pytest.raises(ValueError):
             header12.bps = 5
 
@@ -209,19 +231,19 @@ class TestVDIF:
 
     @pytest.mark.parametrize('edv', [0, 1])  # Others have fixed length
     def test_header_minimal_length(self, edv):
-        for l in range(4):
+        for fl in range(4):
             with pytest.raises(AssertionError):
                 # Less than header length.
-                vdif.VDIFHeader.fromvalues(edv=edv, frame_length=l)
+                vdif.VDIFHeader.fromvalues(edv=edv, frame_length=fl)
 
         header = vdif.VDIFHeader.fromvalues(edv=edv, frame_length=4)
         assert header.payload_nbytes == 0
 
     def test_legacy_header_minimal_length(self):
-        for l in range(2):
+        for fl in range(2):
             with pytest.raises(AssertionError):
                 # Less than header length.
-                vdif.VDIFHeader.fromvalues(edv=False, frame_length=l)
+                vdif.VDIFHeader.fromvalues(edv=False, frame_length=fl)
         header = vdif.VDIFHeader.fromvalues(edv=False, frame_length=2)
         assert header.payload_nbytes == 0
 
@@ -242,7 +264,7 @@ class TestVDIF:
         class VDIFHeaderX(vdif.header.VDIFSampleRateHeader):
             _edv = 0x58
             _header_parser = (vdif.header.VDIFSampleRateHeader._header_parser
-                              + vlbi_base.header.HeaderParser(
+                              | base.header.HeaderParser(
                                   (('nonsense_0', (6, 0, 32, 0x0)),
                                    ('nonsense_1', (7, 0, 8, None)),
                                    ('nonsense_2', (7, 8, 24, 0x1)))))
@@ -257,7 +279,7 @@ class TestVDIF:
             edv=0x58, time=header.time,
             samples_per_frame=header.samples_per_frame,
             station=header.station, sample_rate=header.sample_rate,
-            bps=header.bps, complex_data=header['complex_data'],
+            bps=header.bps, complex_data=header.complex_data,
             thread_id=header['thread_id'], nonsense_0=2000000000,
             nonsense_1=100, nonsense_2=10000000)
 
@@ -286,9 +308,9 @@ class TestVDIF:
         assert headerX['thread_id'] == header['thread_id']
         assert headerX.sample_rate == header.sample_rate
         assert headerX.samples_per_frame == header.samples_per_frame
-        assert headerX.nchan == header.nchan
+        assert headerX.sample_shape == header.sample_shape
         assert headerX.bps == header.bps
-        assert not headerX['complex_data']
+        assert not headerX.complex_data
         assert headerX.mutable is False
         assert headerX['nonsense_0'] == 2000000000
         assert headerX['nonsense_1'] == 100
@@ -296,7 +318,7 @@ class TestVDIF:
 
     def test_decoding(self, tmpdir):
         """Check that look-up levels are consistent with mark5access."""
-        o2h = vlbi_base.encoding.OPTIMAL_2BIT_HIGH
+        o2h = base.encoding.OPTIMAL_2BIT_HIGH
         assert np.all(vdif.payload.lut1bit[0] == -1.)
         assert np.all(vdif.payload.lut1bit[0xff] == 1.)
         assert np.all(vdif.payload.lut1bit.astype(int)
@@ -324,7 +346,7 @@ class TestVDIF:
         assert vdif.VDIFPayload.fromdata(areal, header) == payload1
         # Also check that a circular decode-encode is self-consistent.
         assert vdif.VDIFPayload.fromdata(payload1.data, header) == payload1
-        header['complex_data'] = True
+        header.complex_data = True
         assert vdif.VDIFPayload.fromdata(acmplx, header) == payload2
         assert vdif.VDIFPayload.fromdata(payload2.data, header) == payload2
         # Also check for bps=4.
@@ -339,7 +361,7 @@ class TestVDIF:
                                             payload_nbytes=payload3.nbytes)
         assert vdif.VDIFPayload.fromdata(areal, header) == payload3
         assert vdif.VDIFPayload.fromdata(payload3.data, header) == payload3
-        header['complex_data'] = True
+        header.complex_data = True
         assert vdif.VDIFPayload.fromdata(acmplx, header) == payload4
         assert vdif.VDIFPayload.fromdata(payload4.data, header) == payload4
 
@@ -375,8 +397,7 @@ class TestVDIF:
             # Too few data.
             vdif.VDIFPayload.fromdata(payload[:100], header)
         # check if it works with complex data
-        payload4 = vdif.VDIFPayload(payload.words, nchan=1, bps=2,
-                                    complex_data=True)
+        payload4 = vdif.VDIFPayload(payload.words, bps=2, complex_data=True)
         assert payload4.complex_data is True
         assert payload4.nbytes == 5000
         assert payload4.shape == (10000, 1)
@@ -386,23 +407,26 @@ class TestVDIF:
         with pytest.raises(ValueError):
             vdif.VDIFPayload.fromdata(payload4.data, header)
         header5 = header.copy()
-        header5['complex_data'] = True
+        header5.complex_data = True
         payload5 = vdif.VDIFPayload.fromdata(payload4.data, header5)
         assert payload5 == payload4
         # Check shape for non-power-of-2 bps.  (Note: cannot yet decode.)
-        payload6 = vdif.VDIFPayload(payload.words, nchan=1, bps=7,
-                                    complex_data=False)
+        payload6 = vdif.VDIFPayload(payload.words, bps=7, complex_data=False)
         assert payload6.shape == (1250 * 4, 1)
-        payload7 = vdif.VDIFPayload(payload.words, nchan=1, bps=7,
-                                    complex_data=True)
+        payload7 = vdif.VDIFPayload(payload.words, bps=7, complex_data=True)
         assert payload7.shape == (1250 * 2, 1)
-        payload8 = vdif.VDIFPayload(payload.words, nchan=1, bps=11,
-                                    complex_data=False)
+        payload8 = vdif.VDIFPayload(payload.words, bps=11, complex_data=False)
         assert payload8.shape == (1250 * 2, 1)
-        payload9 = vdif.VDIFPayload(payload.words, nchan=1, bps=11,
-                                    complex_data=True)
+        payload9 = vdif.VDIFPayload(payload.words, bps=11, complex_data=True)
         assert payload9.shape == (1250 * 1, 1)
 
+    def test_invalid_payload(self):
+        with pytest.raises(ValueError, match='multi-channel'):
+            vdif.VDIFPayload(np.zeros(10, '<u4'), sample_shape=(10,), bps=5)
+        # TODO: fix this limitation and replace with actual test!
+        with pytest.raises(ValueError, match='cannot yet'):
+            vdif.VDIFPayload(np.zeros(10, '<u4'), bps=5)
+
     @pytest.mark.parametrize('item', (2, (), -1, slice(1, 3),
                                       slice(2, 4), slice(-3, None)))
     def test_payload_getitem_setitem(self, item):
@@ -496,14 +520,18 @@ class TestVDIF:
         assert len(frameset.frames) == 8
         assert len(frameset) == len(frameset.frames[0])
         assert frameset.samples_per_frame == 20000
-        assert frameset.nchan == 1
+        assert frameset.sample_shape == (8, 1)
         assert frameset.shape == (20000, 8, 1)
         assert frameset.size == 160000
         assert frameset.ndim == 3
         assert frameset.nbytes == 8 * frameset.frames[0].nbytes
+        assert frameset.nchan == 1
+        assert frameset.fill_value == 0.
         assert 'edv' in frameset
         assert 'edv' in frameset.keys()
         assert frameset['edv'] == 3
+        assert bool(frameset['invalid_data']) is False
+        assert frameset.sample_rate == 32. << u.MHz
         # Properties from headers are passed on if they are settable.
         assert frameset.time == frameset.header0.time
         with pytest.raises(AttributeError):
@@ -545,6 +573,15 @@ class TestVDIF:
                                                **frameset4.header0)
         assert frameset6 == frameset4
 
+        frameset6.frames[4].valid = False
+        frameset6.frames[6].sample_rate = frameset6.sample_rate / 2.
+        assert np.all(frameset6.valid == np.array([
+            True, True, True, True, False, True, True, True]))
+        assert np.all(frameset6['invalid_data'] == np.array([
+            False, False, False, False, True, False, False, False]))
+        assert np.all(frameset6.sample_rate == (32. << u.MHz)
+                      * np.array([1., 1., 1., 1., 1., 1., 0.5, 1.]))
+
         with vdif.open(SAMPLE_FILE, 'rb') as fh:
             fh.seek(0)
             # try reading just a few threads
@@ -571,6 +608,7 @@ class TestVDIF:
         assert np.all(frameset[:] == data)
         # Select just samples.
         assert np.all(frameset[15] == data[15])
+        assert np.all(frameset[(16,)] == data[16])
         assert np.all(frameset[10:20] == data[10:20])
         # Select just frames.
         assert np.all(frameset[:, 3] == data[:, 3])
@@ -578,6 +616,10 @@ class TestVDIF:
         # Select just channels (there is only one).
         assert np.all(frameset[:, :, 0] == data[:, :, 0])
         assert np.all(frameset[:, :, :1] == data[:, :, :1])
+        assert np.all(frameset[10, :, 0] == data[10, :, 0])
+        assert np.all(frameset[10, :, :1] == data[10, :, :1])
+        assert np.all(frameset[10, 3, 0] == data[10, 3, 0])
+        assert np.all(frameset[10, 3, :1] == data[10, 3, :1])
         # Check direct access to frames with known results.
         assert np.all(frameset[:12, 0, 0].astype(int)
                       == np.array([-1, -1, 3, -1, 1, -1, 3, -1, 1, 3, -1, 1]))
@@ -593,6 +635,8 @@ class TestVDIF:
         # Select just samples.
         frameset2[15] = -data[15]
         assert np.all(frameset2[15] == -data[15])
+        frameset2[(15,)] = data[15]
+        assert np.all(frameset2[15] == data[15])
         frameset2[10:20] = -1.
         assert np.all(frameset2[10:20] == -1.)
         frameset2[10:20:2] = data[10:20:2]
@@ -609,6 +653,13 @@ class TestVDIF:
         # Select just channels (there is only one).
         frameset2[:, :, 0] = -data[:, :, 0]
         assert np.all(frameset2[:, :, 0] == -data[:, :, 0])
+        frameset2[1, :, 0] = data[1, :, 0]
+        assert np.all(frameset2[1, :, 0] == data[1, :, 0])
+        frameset2[1, 0, 0] = -data[1, 0, 0]
+        assert np.all(frameset2[1, 0, 0] == -data[1, 0, 0])
+        assert np.all(frameset2[1, 1:, 0] == data[1, 1:, 0])
+        assert np.all(frameset2[0, :, 0] == -data[0, :, 0])
+        assert np.all(frameset2[2:, :, 0] == -data[2:, :, 0])
         frameset2[:, :, :1] = 1.
         assert np.all(frameset2[:, :, :1] == 1.)
         # Test header getting/setting.
@@ -898,6 +949,39 @@ class TestVDIF:
             assert fh.tell() == 12
             assert np.all(out_nosqueeze.squeeze() == out_squeeze)
 
+    def test_pickle(self, tmpdir):
+        expected0 = np.array([-1, -1, 3, -1, 1, -1, 3, -1, 1, 3, -1, 1])
+        with vdif.open(SAMPLE_FILE, 'rs') as fh:
+            record = fh.read(6)
+            assert np.all(record[:, 0].astype(int) == expected0[:6])
+            fh.seek(6)
+            pickled = pickle.dumps(fh)
+            with pickle.loads(pickled) as fh2:
+                assert fh2.tell() == 6
+                record2 = fh2.read(6)
+                assert np.all(record2[:, 0].astype(int) == expected0[6:])
+
+            assert fh.tell() == 6
+            record = fh.read(6)
+            assert np.all(record[:, 0].astype(int) == expected0[6:])
+
+        with pickle.loads(pickled) as fh3:
+            assert fh3.tell() == 6
+            fh3.seek(-3, 1)
+            record3 = fh3.read(6)
+            assert np.all(record3[:, 0].astype(int) == expected0[3:9])
+
+        closed = pickle.dumps(fh)
+        with pickle.loads(closed) as fh4:
+            assert fh4.closed
+            with pytest.raises(ValueError):
+                fh4.read(1)
+
+        vdif_file = str(tmpdir.join('simple.vdif'))
+        with vdif.open(vdif_file, 'ws', header0=fh.header0) as fw:
+            with pytest.raises(TypeError):
+                pickle.dumps(fw)
+
     def test_stream_writer(self, tmpdir):
         vdif_file = str(tmpdir.join('simple.vdif'))
         # Try writing a very simple file, using edv=0.  With 16 samples per
@@ -1007,7 +1091,7 @@ class TestVDIF:
         test_file = str(tmpdir.join('test.vdif'))
         with vdif.open(test_file, 'ws', sample_rate=sample_rate,
                        samples_per_frame=samples_per_frame // 8, nthread=1,
-                       nchan=8, complex_data=header0['complex_data'],
+                       nchan=8, complex_data=header0.complex_data,
                        bps=header0.bps, edv=header0.edv,
                        station=header0.station, time=fh.start_time) as fw:
             fw.write(data)
@@ -1030,7 +1114,7 @@ class TestVDIF:
                            -data, -abs(data)]).transpose(1, 2, 0)
         with vdif.open(test_file, 'ws', sample_rate=sample_rate,
                        samples_per_frame=samples_per_frame // 4, nthread=8,
-                       nchan=4, complex_data=header0['complex_data'],
+                       nchan=4, complex_data=header0.complex_data,
                        bps=header0.bps, edv=header0.edv,
                        station=header0.station, time=fh.start_time) as fw:
             fw.write(data4x)
@@ -1123,10 +1207,13 @@ class TestVDIF:
         with vdif.open(SAMPLE_FILE, 'rb') as fh, \
                 open(corrupt_file, 'w+b') as s:
             frameset = fh.read_frameset()
-            header0 = frameset.frames[0].header
-            frameset.tofile(s)
-            frameset = fh.read_frameset()
+            header0 = frameset.frames[0].header.copy()
             frameset.tofile(s)
+            for frame in frameset.frames:
+                frame.header.mutable = True
+            for i in range(5):
+                frameset['frame_nr'] += 1
+                frameset.tofile(s)
             # Now add lots of the final frame, i.e., with the wrong thread_id.
             fh.seek(-5032, 2)
             frame2 = fh.read_frame()
@@ -1182,12 +1269,11 @@ class TestVDIF:
             vdif.open(tmp_file, 's')
 
 
-def test_sequentialfile(tmpdir):
-    """Tests writing and reading of sequential files.
-
-    These tests focus on reading and writing with templates.
-    """
-
+@pytest.mark.parametrize('template,extra_args', [
+    ('f.{file_nr:03d}.vdif', {}),
+    ('{day}.{file_nr:03d}.vdif', {'day': 'f'})])
+def test_template(tmpdir, template, extra_args):
+    """Tests writing and reading of sequential files using a template."""
     # Use sample file as basis of a file sequence.
     with vdif.open(SAMPLE_FILE, 'rs') as fh:
         header = fh.header0.copy()
@@ -1196,14 +1282,13 @@ def test_sequentialfile(tmpdir):
     data = np.concatenate((data, data, data))
 
     # Create a file sequence using template.
-    template = str(tmpdir.join('f.{file_nr:03d}.vdif'))
-    files = sf.FileNameSequencer(template)
-    with vdif.open(files, 'ws', file_size=16*header.frame_nbytes,
-                   nthread=8, **header) as fw:
+    template = str(tmpdir.join(template))
+    with vdif.open(template, 'ws', file_size=16*header.frame_nbytes,
+                   nthread=8, **header, **extra_args) as fw:
         fw.write(data)
 
     # Read in file-sequence and check data consistency.
-    with vdif.open(files, 'rs') as fn:
+    with vdif.open(template, 'rs', **extra_args) as fn:
         assert len(fn.fh_raw.files) == 3
         assert fn.fh_raw.files[-1] == str(tmpdir.join('f.002.vdif'))
         assert fn.header0.time == header.time
@@ -1211,10 +1296,23 @@ def test_sequentialfile(tmpdir):
         assert np.all(data == fn.read())
 
     # Read in one file and check if everything makes sense.
-    with vdif.open(template.format(file_nr=2), 'rs') as fn:
+    with vdif.open(template.format(file_nr=2, **extra_args), 'rs') as fn:
         assert fn.header0.time - header.time - 2. * dtime < 1 * u.ns
         assert np.all(data[80000:] == fn.read())
 
+    if extra_args:
+        # Check that we get proper failure if we do not pass in arguments
+        # required to fill the template.
+        with pytest.raises(KeyError):
+            vdif.open(template, 'rs')
+        with pytest.raises(KeyError):
+            vdif.open(template, 'ws', file_size=16*header.frame_nbytes,
+                      nthread=8, **header)
+
+    # Check that extra arguments are still recognized as problematic.
+    with pytest.raises(TypeError):
+        vdif.open(template, 'rs', walk='silly', **extra_args)
+
 
 def test_vlbi_vdif():
     """Tests SAMPLE_VLBI, which is SAMPLE_VDIF with uncorrected timestamps."""
@@ -1307,16 +1405,17 @@ def test_legacy_vdif(tmpdir):
     assert header['frame_nr'] == 0
     assert header['vdif_version'] == 1
     assert header.nchan == 2
+    assert header.sample_shape == (2,)
     assert header.frame_nbytes == 507 * 8
     assert header.nbytes == 16
-    assert header['complex_data'] is False
+    assert header.complex_data is False
     assert header.bps == 2
     assert header['thread_id'] == 0
     assert header.station == 'AA'
     with open(str(tmpdir.join('test.vdif')), 'w+b') as s:
         header.tofile(s)
         # Add fake payload
-        s.write(np.zeros(503, dtype=np.int64).tostring())
+        s.write(np.zeros(503, dtype=np.int64).tobytes())
         s.seek(0)
         header2 = vdif.VDIFHeader.fromfile(s)
     assert header2 == header
@@ -1352,6 +1451,7 @@ class TestVDIFBPS1:
         assert header0.nchan == 16
         assert header0.bps == 1
         assert header0.samples_per_frame == 4000
+        assert header0.sample_shape == (16,)
 
     def test_stream(self):
         with vdif.open(SAMPLE_BPS1, 'rs', sample_rate=8*u.MHz) as fh:
